@@ -1,4 +1,4 @@
-/* $Id: nss.c 87 2005-11-20 00:45:03Z lennart $ */
+/* $Id: nss.c 105 2007-01-01 16:45:49Z lennart $ */
 
 /***
     This file is part of nss-mdns.
@@ -39,19 +39,32 @@
 #include "avahi.h"
 #endif
 
-#define MAX_ENTRIES 16
-
-#ifdef NSS_IPV4_ONLY
+#if defined(NSS_IPV4_ONLY) && ! defined(MDNS_MINIMAL)
 #define _nss_mdns_gethostbyname2_r _nss_mdns4_gethostbyname2_r
-#define _nss_mdns_gethostbyname_r _nss_mdns4_gethostbyname_r
-#define _nss_mdns_gethostbyaddr_r _nss_mdns4_gethostbyaddr_r
-#elif NSS_IPV6_ONLY
+#define _nss_mdns_gethostbyname_r  _nss_mdns4_gethostbyname_r
+#define _nss_mdns_gethostbyaddr_r  _nss_mdns4_gethostbyaddr_r
+#elif defined(NSS_IPV4_ONLY) && defined(MDNS_MINIMAL)
+#define _nss_mdns_gethostbyname2_r _nss_mdns4_minimal_gethostbyname2_r
+#define _nss_mdns_gethostbyname_r  _nss_mdns4_minimal_gethostbyname_r
+#define _nss_mdns_gethostbyaddr_r  _nss_mdns4_minimal_gethostbyaddr_r
+#elif defined(NSS_IPV6_ONLY) && ! defined(MDNS_MINIMAL)
 #define _nss_mdns_gethostbyname2_r _nss_mdns6_gethostbyname2_r
-#define _nss_mdns_gethostbyname_r _nss_mdns6_gethostbyname_r
-#define _nss_mdns_gethostbyaddr_r _nss_mdns6_gethostbyaddr_r
+#define _nss_mdns_gethostbyname_r  _nss_mdns6_gethostbyname_r
+#define _nss_mdns_gethostbyaddr_r  _nss_mdns6_gethostbyaddr_r
+#elif defined(NSS_IPV6_ONLY) && defined(MDNS_MINIMAL)
+#define _nss_mdns_gethostbyname2_r _nss_mdns6_minimal_gethostbyname2_r
+#define _nss_mdns_gethostbyname_r  _nss_mdns6_minimal_gethostbyname_r
+#define _nss_mdns_gethostbyaddr_r  _nss_mdns6_minimal_gethostbyaddr_r
+#elif defined(MDNS_MINIMAL)
+#define _nss_mdns_gethostbyname2_r _nss_mdns_minimal_gethostbyname2_r
+#define _nss_mdns_gethostbyname_r  _nss_mdns_minimal_gethostbyname_r
+#define _nss_mdns_gethostbyaddr_r  _nss_mdns_minimal_gethostbyaddr_r
 #endif
 
-/* the resolv conf page states that they only support 6 domains */
+/* Maximum number of entries to return */
+#define MAX_ENTRIES 16
+
+/* The resolv.conf page states that they only support 6 domains */
 #define MAX_SEARCH_DOMAINS 6
 
 struct userdata {
@@ -113,46 +126,54 @@ static int ends_with(const char *name, const char* suffix) {
 }
 
 static int verify_name_allowed(const char *name) {
+#ifndef MDNS_MINIMAL
     FILE *f;
-    int valid = 0;
+#endif
     
     assert(name);
 
-    if (!(f = fopen(MDNS_ALLOW_FILE, "r")))
-        return ends_with(name, ".local") || ends_with(name, ".local."); 
-
-    while (!feof(f)) {
-        char ln[128], ln2[128], *t;
+#ifndef MDNS_MINIMAL
+    if ((f = fopen(MDNS_ALLOW_FILE, "r"))) {
+        int valid = 0;
         
-        if (!fgets(ln, sizeof(ln), f))
-            break;
-
-        ln[strcspn(ln, "#\t\n\r ")] = 0;
-
-        if (ln[0] == 0)
-            continue;
-
-        if (strcmp(ln, "*") == 0) {
-            valid = 1;
-            break;
-        }
-
-        if (ln[0] != '.')
-            snprintf(t = ln2, sizeof(ln2), ".%s", ln);
-        else
-            t = ln;
-
-        if (ends_with(name, t)) {
-            valid = 1;
-            break;
+        
+        while (!feof(f)) {
+            char ln[128], ln2[128], *t;
+            
+            if (!fgets(ln, sizeof(ln), f))
+                break;
+            
+            ln[strcspn(ln, "#\t\n\r ")] = 0;
+            
+            if (ln[0] == 0)
+                continue;
+            
+            if (strcmp(ln, "*") == 0) {
+                valid = 1;
+                break;
+            }
+            
+            if (ln[0] != '.')
+                snprintf(t = ln2, sizeof(ln2), ".%s", ln);
+            else
+                t = ln;
+            
+            if (ends_with(name, t)) {
+                valid = 1;
+                break;
+            }
         }
+        
+        fclose(f);
+        return valid;
     }
+#endif
 
-    fclose(f);
-
-    return valid;
+    return ends_with(name, ".local") || ends_with(name, ".local."); 
 }
 
+#ifdef HONOUR_SEARCH_DOMAINS
+
 static char **alloc_domains(unsigned ndomains) {
     char **domains;
 
@@ -257,6 +278,8 @@ static char** get_search_domains(void) {
     return domains;
 }
 
+#endif
+
 enum nss_status _nss_mdns_gethostbyname2_r(
     const char *name,
     int af,
@@ -268,7 +291,7 @@ enum nss_status _nss_mdns_gethostbyname2_r(
 
     struct userdata u;
     enum nss_status status = NSS_STATUS_UNAVAIL;
-    int fd = -1, i;
+    int i;
     size_t address_length, l, idx, astart;
     void (*ipv4_func)(const ipv4_address_t *ipv4, void *userdata);
     void (*ipv6_func)(const ipv6_address_t *ipv6, void *userdata);
@@ -279,8 +302,19 @@ enum nss_status _nss_mdns_gethostbyname2_r(
     uint8_t data[128];
 #endif
 
+#ifdef ENABLE_LEGACY
+    int fd = -1;
+#endif
+
 /*     DEBUG_TRAP; */
 
+    if (af == AF_UNSPEC)
+#ifdef NSS_IPV6_ONLY
+        af = AF_INET6;
+#else
+        af = AF_INET;
+#endif
+
 #ifdef NSS_IPV4_ONLY
     if (af != AF_INET) 
 #elif NSS_IPV6_ONLY
@@ -291,6 +325,7 @@ enum nss_status _nss_mdns_gethostbyname2_r(
     {    
         *errnop = EINVAL;
         *h_errnop = NO_RECOVERY;
+
         goto finish;
     }
 
@@ -302,7 +337,7 @@ enum nss_status _nss_mdns_gethostbyname2_r(
         *errnop = ERANGE;
         *h_errnop = NO_RECOVERY;
         status = NSS_STATUS_TRYAGAIN;
-
+        
         goto finish;
     }
     
@@ -335,10 +370,11 @@ enum nss_status _nss_mdns_gethostbyname2_r(
                 ipv4_func((ipv4_address_t*) data, &u);
             if (af == AF_INET6 && ipv6_func)
                 ipv6_func((ipv6_address_t*)data, &u);
-        }
-        
+        } else
+            status = NSS_STATUS_NOTFOUND;
     }
 
+#ifdef HONOUR_SEARCH_DOMAINS
     if (u.count == 0 && avahi_works && !ends_with(name, ".")) {
         char **domains;
 
@@ -364,15 +400,19 @@ enum nss_status _nss_mdns_gethostbyname2_r(
                     free(fullname);
                     
                     if (r < 0) {
+                        /* Lookup failed */
                         avahi_works = 0;
                         break;
                     } else if (r == 0) {
+                        /* Lookup succeeded */
                         if (af == AF_INET && ipv4_func)
                             ipv4_func((ipv4_address_t*) data, &u);
                         if (af == AF_INET6 && ipv6_func)
                             ipv6_func((ipv6_address_t*)data, &u);
                         break;
-                    }
+                    } else
+                        /* Lookup suceeded, but nothing found */
+                        status = NSS_STATUS_NOTFOUND;
                     
 		} else
                     free(fullname);
@@ -381,21 +421,30 @@ enum nss_status _nss_mdns_gethostbyname2_r(
 	    free_domains(domains);
 	}
     }
+#endif /* HONOUR_SEARCH_DOMAINS */
+#endif /* ENABLE_AVAHI */
 
-
-    if (u.count == 0 && !avahi_works) {
+#if defined(ENABLE_LEGACY) && defined(ENABLE_AVAHI)
+    if (u.count == 0 && !avahi_works) 
 #endif
 
+#if defined(ENABLE_LEGACY)
+    {
         if ((fd = mdns_open_socket()) < 0) {
             *errnop = errno;
             *h_errnop = NO_RECOVERY;
             goto finish;
         }
 
-        if (name_allowed)
+        if (name_allowed) {
             /* Ignore return value */
             mdns_query_name(fd, name, ipv4_func, ipv6_func, &u);
 
+            if (!u.count)
+                status = NSS_STATUS_NOTFOUND;
+        }
+
+#ifdef HONOUR_SEARCH_DOMAINS
         if (u.count == 0 && !ends_with(name, ".")) {
             char **domains;
             
@@ -420,10 +469,12 @@ enum nss_status _nss_mdns_gethostbyname2_r(
                         mdns_query_name(fd, fullname, ipv4_func, ipv6_func, &u);
                         
                         if (u.count > 0) {
-                            /* We found somethings, so let's quit */
+                            /* We found something, so let's quit */
                             free(fullname);
                             break;
-                        }
+                        } else
+                            status = NSS_STATUS_NOTFOUND;
+
                     }
                     
                     free(fullname);
@@ -432,10 +483,9 @@ enum nss_status _nss_mdns_gethostbyname2_r(
                 free_domains(domains);
 	    }
         }
-
-#ifdef ENABLE_AVAHI
+#endif /* HONOUR_SEARCH_DOMAINS */
     }
-#endif
+#endif /* ENABLE_LEGACY */
 
     if (u.count == 0) {
         *errnop = ETIMEDOUT;
@@ -483,9 +533,11 @@ enum nss_status _nss_mdns_gethostbyname2_r(
     status = NSS_STATUS_SUCCESS;
     
 finish:
+#ifdef ENABLE_LEGACY
     if (fd >= 0)
         close(fd);
-    
+#endif
+
     return status;
 }
 
@@ -499,11 +551,7 @@ enum nss_status _nss_mdns_gethostbyname_r (
 
     return _nss_mdns_gethostbyname2_r(
         name,
-#ifdef NSS_IPV6_ONLY
-        AF_INET6,
-#else
-        AF_INET,
-#endif        
+        AF_UNSPEC,
         result,
         buffer,
         buflen,
@@ -523,21 +571,26 @@ enum nss_status _nss_mdns_gethostbyaddr_r(
     
     struct userdata u;
     enum nss_status status = NSS_STATUS_UNAVAIL;
-    int fd = -1, r;
+    int r;
     size_t address_length, idx, astart;
+    
 #ifdef ENABLE_AVAHI
     char t[256];
 #endif
+#ifdef ENABLE_LEGACY
+    int fd = -1;
+#endif
 
     *errnop = EINVAL;
     *h_errnop = NO_RECOVERY;
 
     u.count = 0;
     u.data_len = 0;
-    
+
+    /* Check for address types */
     address_length = af == AF_INET ? sizeof(ipv4_address_t) : sizeof(ipv6_address_t);
 
-    if (len != (int) address_length ||
+    if (len < (int) address_length ||
 #ifdef NSS_IPV4_ONLY
         af != AF_INET
 #elif NSS_IPV6_ONLY
@@ -548,9 +601,11 @@ enum nss_status _nss_mdns_gethostbyaddr_r(
         ) {
         *errnop = EINVAL;
         *h_errnop = NO_RECOVERY;
+
         goto finish;
     }
 
+    /* Check for buffer space */
     if (buflen <
         sizeof(char*)+      /* alias names */
         address_length) {   /* address */
@@ -562,53 +617,81 @@ enum nss_status _nss_mdns_gethostbyaddr_r(
         goto finish;
     }
 
-#ifdef ENABLE_AVAHI
+#ifdef MDNS_MINIMAL
 
+    /* Only query for 169.254.0.0/16 IPv4 in minimal mode */
+    if ((af == AF_INET && ((ntohl(*(const uint32_t*)  addr) & 0xFFFF0000UL) != 0xA9FE0000UL)) ||
+        (af == AF_INET6 && !(((const uint8_t*) addr)[0] == 0xFE && (((const uint8_t*) addr)[1] >> 6) == 2))) {
+
+        *errnop = EINVAL;
+        *h_errnop = NO_RECOVERY;
+
+        goto finish;
+    }
+#endif
+    
+#ifdef ENABLE_AVAHI
+    /* Lookup using Avahi */
     if ((r = avahi_resolve_address(af, addr, t, sizeof(t))) == 0) {
         name_callback(t, &u);
     } else if (r > 0) {
         *errnop = ETIMEDOUT;
         *h_errnop = HOST_NOT_FOUND;
+        status = NSS_STATUS_NOTFOUND;
         goto finish;
-    }
-
-#endif 
-
-    if (u.count == 0) {
+    } 
+#endif
 
+#if defined(ENABLE_AVAHI) && defined(ENABLE_LEGACY)
+    else
+#endif
+    
+#ifdef ENABLE_LEGACY
+     /* Lookup using legacy mDNS queries */   
+     {
         if ((fd = mdns_open_socket()) < 0) {
-            
             *errnop = errno;
             *h_errnop = NO_RECOVERY;
             goto finish;
         }
-        
+
+	r = -1;
+
 #if ! defined(NSS_IPV6_ONLY) && ! defined(NSS_IPV4_ONLY)
         if (af == AF_INET)
 #endif
-#ifndef NSS_IPV6_ONLY        
+#ifndef NSS_IPV6_ONLY
             r = mdns_query_ipv4(fd, (const ipv4_address_t*) addr, name_callback, &u);
 #endif
 #if ! defined(NSS_IPV6_ONLY) && ! defined(NSS_IPV4_ONLY)
         else
 #endif
-#ifndef NSS_IPV4_ONLY        
+#ifndef NSS_IPV4_ONLY
             r = mdns_query_ipv6(fd, (const ipv6_address_t*) addr, name_callback, &u);
 #endif
-        
         if (r < 0) {
             *errnop = ETIMEDOUT;
             *h_errnop = HOST_NOT_FOUND;
+            status = NSS_STATUS_NOTFOUND;
             goto finish;
         }
     }
+#endif /* ENABLE_LEGACY */
+
+    if (u.count == 0) {
+        *errnop = ETIMEDOUT;
+        *h_errnop = NO_RECOVERY;
+        goto finish;
+    }
 
     /* Alias names */
     *((char**) buffer) = NULL;
     result->h_aliases = (char**) buffer;
     idx = sizeof(char*);
 
-    assert(u.count > 0 && u.data.name[0]);
+    assert(u.count > 0);
+    assert(u.data.name[0]);
+    
     if (buflen <
         strlen(u.data.name[0])+1+ /* official names */
         sizeof(char*)+ /* alias names */
@@ -642,9 +725,11 @@ enum nss_status _nss_mdns_gethostbyaddr_r(
     status = NSS_STATUS_SUCCESS;
     
 finish:
+#ifdef ENABLE_LEGACY
     if (fd >= 0)
         close(fd);
-    
+#endif
+
     return status;
 }
 
